local Addon, private = ...

-- Builtins
local assert = assert
local math = math
local pairs = pairs
local print = print
local select = select
local string = string
local table = table

-- Globals
local _G = _G
local Command = Command
local dump = dump
local Event = Event
local Inspect = Inspect
local UI = UI
local UIParent = UIParent
local Utility = Utility

-- Locals
local padding = 5
local scrollbarWidth = 16
local labelFormats = {
	background = "Background",
	size = "%d x %d",
	randomNames = "Randomized Names",
}
local defaultSettings = {
	background = false,
	randomNames = true,
}

-- Prototypes
local textureSelected

setfenv(1, private)

if(Addon.toc.debug) then
	log = print
else
	log = function() end
end

-- UI creation methods
-- ============================================================================

local function makeMovable(self)
	local mouseOffsetX
	local mouseOffsetY
	local function down()
		local mouse = Inspect.Mouse()
		mouseOffsetX, mouseOffsetY = self:GetLeft() - mouse.x, self:GetTop() - mouse.y
	end
	local function up()
		mouseOffsetX, mouseOffsetY = nil, nil
	end
	local function move()
		if(mouseOffsetX) then
			local mouse = Inspect.Mouse()
			local x = mouse.x + mouseOffsetX
			local y = mouse.y + mouseOffsetY
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
		end
	end
	local content = self:GetContent()
	content:EventAttach(Event.UI.Input.Mouse.Left.Down, down, "")
	content:EventAttach(Event.UI.Input.Mouse.Left.Up, up, "")
	content:EventAttach(Event.UI.Input.Mouse.Left.Upoutside, up, "")
	content:EventAttach(Event.UI.Input.Mouse.Cursor.Move, move, "")
	local border = self:GetBorder()
	border:EventAttach(Event.UI.Input.Mouse.Left.Down, down, "")
	border:EventAttach(Event.UI.Input.Mouse.Left.Up, up, "")
	border:EventAttach(Event.UI.Input.Mouse.Left.Upoutside, up, "")
	border:EventAttach(Event.UI.Input.Mouse.Cursor.Move, move, "")
end

local function createTextureList(parent, width, entries)
	local self = UI.CreateFrame("Frame", "", parent)
	self:SetBackgroundColor(0, 0, 0)
	self.names = { }
	self.paths = { }
	local selected = 0
	
	local scrollbar = UI.CreateFrame("RiftScrollbar", "", self)
	scrollbar:SetPoint("TOPLEFT", self, "TOPRIGHT", padding, 0)
	scrollbar:SetPoint("BOTTOMLEFT", self, "BOTTOMRIGHT", padding, 0)
	scrollbar:EventAttach(Event.UI.Scrollbar.Change, function() self:Update() end, "")
	
	self:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function() scrollbar:NudgeDown() end, "")
	self:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function() scrollbar:NudgeUp() end, "")
	
	local function index(i)
		return math.floor(scrollbar:GetPosition()) + i
	end
	
	local function keyDown(self, h, key)
		if(key == "Down") then
			selected = math.min(selected + 1, #self.paths)
		elseif(key == "Up") then
			selected = math.max(2, selected - 1) -- Skip "Recorded"
		else
			return
		end
		if(selected < index(1)) then
			scrollbar:NudgeUp()
		elseif(selected > index(entries)) then
			scrollbar:NudgeDown()
		else
			self:Update()
		end
		textureSelected(parent, self.paths[selected])
	end
	self:EventAttach(Event.UI.Input.Mouse.Left.Down, function() self:SetKeyFocus(true) end, "")
	self:EventAttach(Event.UI.Input.Key.Down, keyDown, "")
	self:EventAttach(Event.UI.Input.Key.Repeat, keyDown, "")
	
	local function label_MouseIn(label)
		label:SetBackgroundColor(0.3, 0.3, 0.3)
	end
	local function label_MouseOut(label)
		if(index(label.offset) ~= selected) then
			label:SetBackgroundColor(0, 0, 0)
		end
	end
	local function label_LeftDown(label)
		self:SetKeyFocus(true)
		if(label.path ~= "") then
			selected = index(label.offset)
			textureSelected(parent, label.path)
			self:Update()
		end
	end
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetText("X")
	label.offset = 1
	label:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
	label:SetPoint("TOPRIGHT", scrollbar, "TOPLEFT", -padding, 0)
	label:EventAttach(Event.UI.Input.Mouse.Cursor.In, label_MouseIn, "")
	label:EventAttach(Event.UI.Input.Mouse.Cursor.Out, label_MouseOut, "")
	label:EventAttach(Event.UI.Input.Mouse.Left.Down, label_LeftDown, "")
	self.labels = { label }
	
	for i = 2, entries do
		label = UI.CreateFrame("Text", "", self)
		label.offset = i
		label:SetText("X")
		label:SetPoint("TOPLEFT", self.labels[i - 1], "BOTTOMLEFT")
		label:SetPoint("TOPRIGHT", self.labels[i - 1], "BOTTOMRIGHT")
		label:EventAttach(Event.UI.Input.Mouse.Cursor.In, label_MouseIn, "")
		label:EventAttach(Event.UI.Input.Mouse.Cursor.Out, label_MouseOut, "")
		label:EventAttach(Event.UI.Input.Mouse.Left.Down, label_LeftDown, "")
		self.labels[i] = label
	end
	
	self:SetHeight(self.labels[1]:GetHeight() * entries)
	self:SetWidth(width)
	
	function self:SetTextures(names, paths)
		assert(#names == #paths)
		self.names = names
		self.paths = paths
		scrollbar:SetRange(0, math.max(1, #names - entries))
		scrollbar:SetEnabled(#names > entries)
		scrollbar:SetPosition(0)
		selected = 0
		self:Update()
	end
	
	function self:Update()
		local offset = math.floor(scrollbar:GetPosition())
		for i = 1, entries do
			local index = i + offset
			if(index > #self.names) then
				self.labels[i]:SetVisible(false)
			else
				self.labels[i]:SetVisible(true)
				self.labels[i]:SetText(self.names[index])
				self.labels[i].path = self.paths[index]
				if(selected == index) then
					self.labels[i]:SetBackgroundColor(0.3, 0.3, 0.3)
				else
					self.labels[i]:SetBackgroundColor(0, 0, 0)
				end
			end
		end
	end
	
	return self
end

local function createTextureView(parent)
	local self = UI.CreateFrame("Texture", "", parent)
	self:SetTexture("TextureViewer", "checkerboard.png")
	
	self.width = 0
	self.height = 0
	self.zoom = 1.0
	self:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function() self:SetZoom(self.zoom * 2) end, "")
	self:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function() self:SetZoom(self.zoom / 2) end, "")
	
	local scrollbar_v = UI.CreateFrame("RiftScrollbar", "", self)
	scrollbar_v:SetPoint("TOPLEFT", self, "TOPRIGHT", 0, 0)
	scrollbar_v:SetPoint("BOTTOMLEFT", self, "BOTTOMRIGHT", 0, 0)
	scrollbar_v:EventAttach(Event.UI.Scrollbar.Change, function() self:Update() end, "")

	local scrollbar_h = UI.CreateFrame("RiftScrollbar", "", parent)
	scrollbar_h:SetParent(self)
	scrollbar_h:SetOrientation("horizontal")
	scrollbar_h:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 0, 0)
	scrollbar_h:SetPoint("TOPRIGHT", self, "BOTTOMRIGHT", 0, 0)
	scrollbar_h:EventAttach(Event.UI.Scrollbar.Change, function() self:Update() end, "")

	self.mask = UI.CreateFrame("Mask", "", self)
	self.mask:SetAllPoints(self)
	
	self.texture = UI.CreateFrame("Texture", "", self.mask)
	self.texture:SetPoint("CENTER", self.mask, "CENTER")
	self.texture:SetWidth(0)
	self.texture:SetHeight(0)
	
	self.texture:EventAttach(Event.UI.Input.Mouse.Left.Down, function(self)
		self.down = Inspect.Mouse()
	end, "")
	self.texture:EventAttach(Event.UI.Input.Mouse.Left.Up, function(self)
		self.down = nil
	end, "")
	self.texture:EventAttach(Event.UI.Input.Mouse.Left.Upoutside, function(self)
		self.down = nil
	end, "")
	
	self.texture:EventAttach(Event.UI.Input.Mouse.Cursor.Move, function()
		if(self.texture.down) then
			local down = Inspect.Mouse()
			local range_h, range_v = select(2, scrollbar_h:GetRange()), select(2, scrollbar_v:GetRange())
			scrollbar_h:SetPosition(math.max(0, math.min(range_h, scrollbar_h:GetPosition() + (self.texture.down.x - down.x))))
			scrollbar_v:SetPosition(math.max(0, math.min(range_v, scrollbar_v:GetPosition() + (self.texture.down.y - down.y))))
			self.texture.down = down
		end
	end, "")
	
	function self:Update()
		local width, height = self.texture:GetWidth(), self.texture:GetHeight()
		local range_h, range_v = select(2, scrollbar_h:GetRange()), select(2, scrollbar_v:GetRange())
		self.texture:SetPoint("CENTER", self.mask, "CENTER", range_h / 2 - scrollbar_h:GetPosition(), range_v / 2 - scrollbar_v:GetPosition())
	end
	
	function self:SetZoom(factor)
		self.zoom = factor
		local width, height = self.width * self.zoom, self.height * self.zoom
		self.texture:SetWidth(width)
		self.texture:SetHeight(height)
		if(width <= self:GetWidth()) then
			scrollbar_h:SetEnabled(false)
			scrollbar_h:SetRange(0, 0)
		else
			scrollbar_h:SetEnabled(true)
			scrollbar_h:SetRange(0, width - self:GetWidth())
			scrollbar_h:SetThickness(self:GetWidth())
			scrollbar_h:SetPosition(select(2, scrollbar_h:GetRange()) / 2)
		end
		
		if(height <= self:GetHeight()) then
			scrollbar_v:SetEnabled(false)
			scrollbar_v:SetRange(0, 0)
		else
			scrollbar_v:SetEnabled(true)
			scrollbar_v:SetRange(0, height - self:GetHeight())
			scrollbar_v:SetThickness(self:GetHeight())
			scrollbar_v:SetPosition(select(2, scrollbar_v:GetRange()) / 2)
		end
		self:Update()
	end
	
	function self:SetTexture(source, texture)
		self.zoom = 1.0
		if(texture == "") then
			self.texture:SetWidth(0)
			self.texture:SetHeight(0)
		else
			self.texture:ClearWidth()
			self.texture:ClearHeight()
		end
		self.texture:SetTexture(source, texture)
		self.width, self.height = self.texture:GetWidth(), self.texture:GetHeight()
		self:SetZoom(self.zoom)
	end

	function self:SetBackground(opaque)
		if(opaque) then
			self.texture:SetBackgroundColor(0, 0, 0, 1)
		else
			self.texture:SetBackgroundColor(0, 0, 0, 0)
		end
	end
	
	function self:GetTextureSize()
		return self.texture:GetWidth(), self.texture:GetHeight()
	end
	
	return self
end

local function createCheckbox(parent, text, callback)
	local self = UI.CreateFrame("RiftCheckbox", "", parent)
	
	self.text = UI.CreateFrame("Text", "", self)
	self.text:SetPoint("LEFTCENTER", self, "RIGHTCENTER")
	self.text:SetText(text)
	
	self:EventAttach(Event.UI.Checkbox.Change, callback, "")
	
	return self
end

local function setTexture(self, source, texture)
	self.textureView:SetTexture(source, texture)
	self.sizeLabel:SetText(string.format(labelFormats.size, self.textureView:GetTextureSize()))
	self.texturePath:SetText(string.format("%q", texture))
end

local function createWindow()
	Command.System.Watchdog.Quiet()
	
	local context = UI.CreateContext(Addon.identifier)
	local self = UI.CreateFrame("RiftWindow", "", context)
	self:SetTitle(Addon.toc.Name)
	self.data = ""
	
	self:SetController("content")
	self:SetWidth(935)
	self:SetHeight(565)
	self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (UIParent:GetWidth() - self:GetWidth()) / 2, (UIParent:GetHeight() - self:GetHeight()) / 2)
	
	local btn = UI.CreateFrame("RiftButton", "", self)
	btn:SetSkin("close")
	btn:SetPoint("TOPRIGHT", self:GetBorder(), "TOPRIGHT", -8, 16)
	btn:EventAttach(Event.UI.Button.Left.Press, function()
		self:SetVisible(false)
		self.textureList:SetKeyFocus(false)
	end, "")
	
	-- Texture list
	self.textureList = createTextureList(self, 206, 29)
	self.textureList:SetPoint("TOPLEFT", self, "TOPLEFT", padding, padding)
	self.textureList:SetTextures(TextureNames, TexturePaths)
	
	-- Size text
	self.sizeLabel = UI.CreateFrame("Text", "", self)
	self.sizeLabel:SetPoint("TOPRIGHT", self, "TOPRIGHT", -2 * padding - scrollbarWidth, 0)
	self.sizeLabel:SetText("")
	
	-- Texture path
	self.texturePath = UI.CreateFrame("RiftTextfield", "", self)
	self.texturePath:SetPoint("TOPLEFT", self.textureList, "TOPRIGHT", 2 * padding + scrollbarWidth, 0)
	self.texturePath:SetPoint("TOPRIGHT", self.sizeLabel, "TOPLEFT", -padding, 0)
	self.texturePath:SetText("")
	
	-- Texture view
	self.textureView = createTextureView(self)
	self.textureView:SetPoint("TOPLEFT", self.texturePath, "BOTTOMLEFT", 0, 0)
	self.textureView:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -2 * padding - scrollbarWidth, -2 * padding - 2 * scrollbarWidth)

	-- Background checkbox
	self.backgroundCheck = createCheckbox(self, labelFormats.background, function(cb)
		self.textureView:SetBackground(cb:GetChecked())
		_G.TextureViewerSettings.background = cb:GetChecked()
	end)
	self.backgroundCheck:SetPoint("BOTTOMLEFT", self.textureList, "BOTTOMRIGHT", 2 * padding + scrollbarWidth, 0)
	self.backgroundCheck:SetChecked(_G.TextureViewerSettings.background)
	
	-- Random names checkbox
	self.randomNamesCheck = createCheckbox(self, labelFormats.randomNames, function(cb)
		_G.TextureViewerSettings.randomNames = cb:GetChecked()
	end)
	self.randomNamesCheck:SetPoint("LEFTCENTER", self.backgroundCheck, "RIGHTCENTER", 100, 0)
	self.randomNamesCheck:SetChecked(_G.TextureViewerSettings.randomNames)
	
	self.SetTexture = setTexture
	self:SetTexture("Rift", "")
	
	makeMovable(self)
	MainWindow = self
end

textureSelected = function(self, path)
	setTexture(self, "Rift", path)
end

-- Message processing methods
-- ============================================================================

local function toggleWindow()
	if(not MainWindow) then
		createWindow()
	else
		MainWindow:SetVisible(not MainWindow:GetVisible())
	end
end

local function variablesLoaded(identifier)
	if(identifier ~= Addon.identifier) then
		return
	end
	
	_G.TextureViewerSettings = _G.TextureViewerSettings or defaultSettings
	local sorted = { }
	for k, v in pairs(_G.TextureViewerRecordedTextures or { }) do
		if(_G.TextureViewerSettings.randomNames or not string.match(k, "^%a+_I%x+%.dds$")) then
			sorted[#sorted + 1] = k
		end
	end
	
	table.sort(sorted)
	for k, v in pairs(sorted) do
		TextureNames[#TextureNames + 1] = "\t" .. v
		TexturePaths[#TexturePaths + 1] = v
	end
end

local function recordTextures(textures)
	_G.TextureViewerRecordedTextures = _G.TextureViewerRecordedTextures or { }
	for k in pairs(textures) do
		if(not Blacklist[k] and k ~= "") then
			_G.TextureViewerRecordedTextures[k] = true
		end
	end
end

if(Command.System.Texture) then
	Command.System.Texture.Record()
end

table.insert(Command.Slash.Register(Addon.identifier), { toggleWindow, Addon.identifier, "toggleWindow" })
table.insert(Command.Slash.Register("tv"), { toggleWindow, Addon.identifier, "toggleWindow" })

table.insert(Event.Addon.SavedVariables.Load.End, { variablesLoaded, Addon.identifier, "variablesLoaded" })
if(Event.System.Texture) then
	table.insert(Event.System.Texture, { recordTextures, Addon.identifier, "recordTextures" })
end
--toggleWindow()
